/*
 * lk - a lightweight terminal client over serial line
 *
 * Copyright (c) 2003-2006 Viliam Holub
 *
 * lk is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * lk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with tinypcb; see the file COPYING.
 * If not, write to the Free Software Foundation,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <sys/types.h>
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>

#include "cross.h"
#include "parse.h"
#include "fault.h"
#include "termf.h"
#include "text.h"
#include "safecall.h"



#define BREAK_CHAR	0x1c



/** Investigate program name.
 */
void
init_prog_name( const char *arg)

{
	int c=0;
	const char *oarg = arg;
	
	for (;*arg;arg++,c++)
		if (*arg == '/')
		{
			oarg = arg;
			c = 0;
		}
	
	if (c<2)
	{
		txt_progname = xstrdup( txt_default_progname);
		return;
	}
	
	txt_progname = xmalloc( c);
	strcpy( txt_progname, oarg+1);
}


/** Safe character reading.
 */
char
read_char( int fd)

{
	char c;
	s_read( fd, &c, 1);
	return c;
}


/** Safe character writting.
 */
void
write_char( int fd, char c)

{
	s_write( fd, &c, 1);
}
			

/** User request.
 */
bool
user_request( void)
	
{
	char c = read_char( 0); 

	switch (c)
	{
		case BREAK_CHAR:
			write_char( line, BREAK_CHAR);
			break;
			
		case 'q': /* quit */
			return true;
			
		case 'd': /* detach */
			line_hangup( DONT_RECOVER);
			return true;

		case 'h': /* hangup */
			line_hangup( DO_RECOVER);
			break;
			
		case 'b': /* serial break */
			line_break();
			break;
			
		case '?': /* help */
			fprintf( stderr, txt_screen_help);
			break;
			
		default: /* ignore */
			break;
	}
	
	return false;
}


/** This is the main I/O loop.
 */
void
big_loop( void)

{
	char c;
	int ret;
	fd_set fdsr;
	struct timeval tv;
	
	while (1) 
	{
		FD_ZERO( &fdsr);
		FD_SET( 0, &fdsr);
		FD_SET( line, &fdsr);
		
		tv.tv_usec = 0;
		tv.tv_sec = 35;
		
		ret = select( 16, &fdsr, 0, 0, &tv);
		
		if (FD_ISSET( line, &fdsr))
			write_char( 1, read_char( line));
		
		if (FD_ISSET( 0, &fdsr))
		{
			c = read_char( 0);
			
			if (c == BREAK_CHAR)
			{
				if (user_request())
					break;
			}
			else
				write_char( line, c);
		}
	}
	
	printf( txt_CRNL);
}


/** Signal handler - a request to terminate.
 *
 * What to do now? Close the line, get everything back a exit.
 */
void
terminate( int i)

{
	term_done();
	line_done();
	exit( 0);
}


/** here we have the main control over the program.
 */
int
main( int argc, const char **args)

{
	init_prog_name( *args);

	/* Argument passing. */
	parse_param( args);

	/* Argument dump. */
	if (verbose)
	{
		fprintf( stderr, "Recognized parameters:\n"
				"\tcarrier_control: %d\n"
				"\tflow control input: %d\n"
				"\tflow control output: %d\n"
				"\tdevice name: %s\n"
				"\tbaud_rate: %s\n"
				"\tcharacter size: %c\n"
				"\tparity: %d\n",
				carrier_control,
				flow_control_input,
				flow_control_output,
				device_name,
				baud_rate_set ? baud_rates[ baud_rate].info :
						"undefined",
				character_size == -1 ? 'x' : character_size,
				parity);
	}

	/* Inicializing. */
	term_init();
	line_init();

	signal( SIGINT, SIG_IGN);
	signal( SIGQUIT, SIG_IGN);
	signal( SIGHUP, terminate);
 	signal( SIGUSR1, terminate); 
	signal( SIGUSR2, terminate); 
	signal( SIGTERM, terminate);
	signal( SIGPIPE, terminate);

	/* Show info. */
	info( txt_enter_raw);

	/* Main loop */
	big_loop();

	/* Restoring. */
	term_done();
	line_done();

	return RET_OK;
}
