/*
 * Copyright (c) 2003-2006 Viliam Holub
 */

#include <unistd.h>
#include <termios.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <stdio.h>
#include <sys/types.h>
#include <pwd.h>

#include "cross.h"
#include "parse.h"
#include "fault.h"
#include "safecall.h"
#include "text.h"


/* Maximum line length in configuration file. */
#define CONF_LINE_MAX	1024


/*
 * line control variables
 */
tbool carrier_control = undef;
tbool flow_control_input = undef;
tbool flow_control_output = undef;
tbool parity_check = undef;
tbool double_stop = undef;
tbool ixon = undef;
tbool ixoff = undef;

signed char character_size = -1;
bool baud_rate_set = false;
int baud_rate = 0;
int parity = -1;

int verbose = 0;

const char *device_path = "/dev/";
const char *device_std = "/dev/ttyS0";
char *device_name;


/*
 * Miscellaneous flags
 */
bool pdef = undef;


static void parm_help( char sw);
static void parm_verbose( char sw);
static void parm_flow_control( char sw);
static void parm_ix( char sw);


static void param( const char *par);


/*
 * switchable parameters
 * first we try to upgrade variable, then try to call function
 * <name>, <variable>, <function to handle>
 */
const struct {char *s; bool *b; void (*f)(char sw);} params[] =
{
	{ "-help",	0,			parm_help},
	{ "help",	0,			parm_help},
	{ "h",		0,			parm_help},

	{ "v",		0, 			parm_verbose},
	{ "-verbose",	0, 			parm_verbose},
	
	{ "cc",		&carrier_control,	0},
	{ "fc",		&flow_control_input,	parm_flow_control},
	{ "fci",	&flow_control_input,	0},
	{ "fco",	&flow_control_output,	0},
	{ "pc",		&parity_check,		0},
	{ "ds",		&double_stop,		0},
	{ "ix",		&ixon,			parm_ix},
	{ "ixon",	&ixon,			0},
	{ "ixoff",	&ixoff,			0},

	{ 0, 0}
};


/** Baud rates constants
 * <text value>, <constant>, <valid>, <name>
 */
const struct baud_rate_struct baud_rates[] = 
{
	{ "0",		-1,		false, 		"0"},
	{ "50",		B50,		true,		"50"},
	{ "75",		B75,		true,		"75"},
	{ "110",	B110,		true,		"110"},
	{ "134",	B134,		true,		"134"},
	{ "150",	B150,		true,		"150"},
	{ "200",	B200,		true,		"200"},
	{ "300",	B300,		true,		"300"},
	{ "600",	B600,		true,		"600"},
	{ "1200",	B1200,		true,		"1k2"},
	{ "1800",	B1800,		true,		"1k8"},
	{ "2400",	B2400,		true,		"2k4"},
	{ "4800",	B4800,		true,		"4k8"},
	{ "9600",	B9600,		true,		"9k6"},
	{ "19200",	B19200,		true,		"19k2"},
	{ "38400",	B38400,		true,		"38k4"},
	{ "57600",	B57600,		CB57600,	"57k6"},
	{ "76800",	B76800,		CB76800,	"76k8"},
	{ "115200",	B115200,	CB115200,	"115k2"},
	{ "153600",	B153600,	CB153600,	"153k6"},
	{ "230400",	B230400,	CB230400,	"230k4"},
	{ "307200",	B307200,	CB307200,	"307k2"},
	{ "460800",	B460800,	CB460800,	"460k8"},
	{ "500000",	B500000,	CB500000,	"500k"},
	{ "576000",	B576000,	CB576000,	"576k"},
	{ "921600",	B921600,	CB921600,	"921k6"},
	{ "1000000",	B1000000,	CB1000000,	"1M"},
	{ "1152000",	B1152000,	CB1152000,	"1M152"},
	{ "1500000",	B1500000,	CB1500000,	"1M5"},
	{ "2000000",	B2000000,	CB2000000,	"2M"},
	{ "2500000",	B2500000,	CB2500000,	"2M5"},
	{ "3000000",	B3000000,	CB3000000,	"3M"},
	{ "3500000",	B3500000,	CB3500000,	"3M5"},
	{ "4000000",	B4000000,	CB4000000,	"4M"},
	{ 0, 0, 0}
};


/** Configuration functions
 *
 * These functions are used to parse switchable parameters which are not
 * easy to handle.
 */


/** Shows the help message and exits.
 */
static void
parm_help( char sw)

{
	fprintf( stderr, txt_help);
	exit( 0);
	/* unreachable */
}


/** Verbose parameter.
 */
static void
parm_verbose( char sw)

{
	verbose = 1;
}


/** Only update flow_control_output, input is already configured.
 */
static void
parm_flow_control( char sw)

{
	flow_control_output = flow_control_input;
}


/* Only update ixoff, ixon is already configured.
 */
static void
parm_ix( char sw)

{
	ixoff = ixon;
}



/** Partial string compare;.
 * 
 * on fail returns false
 * on succes s2 points to the first different char; returns true
 */
static bool
pstrcmp( const char *s1, const char **ss2)

{
	bool res;
	const char *s2 = *ss2;
	
	while (*s1 && (*s1 == *s2))
	{
		s1++;
		s2++;
	}

	if ((res = !*s1 && !isdigit( *s2)))
		*ss2 = s2;

	return res;
}


/** Converts the switchable parameter to a real variable value.
 *
 * par	option must begin with '+', '-' or '='
 */
static void
param_switch( const char *par)

{
	int i=0;
	
	/* Find option index. */
	while (params[ i].s && strcmp( params[ i].s, par+1))
		i++;
	
	/* Parameter not found. */
	if (!params[ i].s) 
		die_parm( par, txt_invpar);
	
	/* Action. */
	if (params[ i].b)
		/* Set specified variable. */
		switch (*par)
		{
			case '+' : *params[ i].b = true; break;
			case '-' : *params[ i].b = false; break;
			case '=' : *params[ i].b = undef; break;
		}

	/* Call appropriate handler if defined. */
	if (params[ i].f)
		params[ i].f( *par);
}


/** Get a device name from the parameter.
 *
 * The parameter is copied into device_name variable. A parameter
 * is updated to point below the device name.
 * Also checks device name size and die on fault
 */
static void
param_str_device( const char **par)

{
	const char *s = *par;
	const char *opt = *par;
	char *c;
	int len=0;
	size_t device_path_len = strlen( device_path);
	
	/* If the device name is not defined. */
	if (*s == ',')
	{
		(*par)++;
		return;
	}
	
	/* Search for the end of the device name. */
	while (*s && (*s != ','))
	{
		len++;
		s++;
	}

	/* Create a new device name space. */
	if (device_name != device_std)
		free( device_name);
	device_name = c = (char *)xmalloc( 1 +len
			+((*opt == '/' ? 0 : device_path_len)));
	
	/* Copy device name. */
	if (*opt != '/')
	{
		strcpy( c, device_path);
		c += device_path_len;
	}
	memcpy( c, opt, len);

	*(c+len) = '\0';
	*par += len;
}


/** Handles a param which starts on numeric value.
 */
static void
param_str_num( const char **par)

{
	int i;
	
	if (!**par)
		return;
	
	if (**par == ',')
		(*par)++;
	
	if (!isdigit( **par))
		die_parm( *par, txt_illbaud);
	
	i=0;
	while (baud_rates[ i].s && (!pstrcmp( baud_rates[ i].s, par)))
		i++;
	
	if (!baud_rates[ i].s || !baud_rates[ i].c)
		die_parm( *par, txt_illbaud2);
	
	if (i != 0)
	{
		baud_rate_set = true;
		baud_rate = baud_rates[ i].br;
	}
}


#define CHAR2STR( c)	char msg[ 2]; msg[ 0] = c; msg[ 1] = 0;


/** Handles parity parameter.
 */
static void
param_str_parity( const char **p)

{
	if (!**p)
		return;

	switch (**p)
	{
		case 'n':
		case 'N':
			parity = 0;
			break;
		case 'e':
		case 'E':
			parity = PARENB;
			break;
		case 'o':
		case 'O':
			parity = PARENB|PARODD;
			break;
		default:
			{
				CHAR2STR( **p)
				die_parm( msg, txt_illparity);
			}
	}

	(*p)++;
}


/** Handles character size in parameter.
 */
static void
param_str_char_size( const char **p)

{
	if (!**p)
		return;

	if (**p >= '5' && **p <= '8')
		character_size = **p;
	else
	{
		CHAR2STR( **p);
		die_parm( msg, txt_illchar);
	}
	
	(*p)++;
}


/** Process string parameter.
 *
 * par must be in the following format:
 * device,speed_parity_bits
 * 
 * example: ttyS1,9600n7
 */
static void
param_str( const char *par)

{
	if (verbose)
		info( "Interpreting \"%s\" as an in-line configuration.", par);

	if (!isdigit( *par))
		param_str_device( &par);
	param_str_num( &par);
	param_str_parity( &par);
	param_str_char_size( &par);

	if (*par)
		die_parm( par, txt_invpar);
}


/** Returns the user home directory.
 *
 * We get a home directory from the passwd file via getpw*. First try to
 * get user name specified by the LOGNAME environment variable. If unset, we
 * index users via uid.
 */
static const char *
get_user_home( void)

{
	struct passwd *pw = NULL;
	char *logname = getenv( "LOGNAME");
	if (logname)
		pw = getpwnam( logname);
	
	if (!pw)
		pw = getpwuid( getuid());

	return pw ? pw->pw_dir : "./";
}


/* Try to find the name in the configuration file.
 */
static bool
param_find( const char *name)

{
	FILE *f;
	static int recursive = 0;
	const char *home = get_user_home();
	size_t home_size = strlen( home);
	size_t conf_size = strlen( conf_name);
	char *configuration;

	/* Check recursive calls. */
	if (++recursive == 100)
		die( RET_PAR, "Too deep recursive definition.");

	/* Create full path for configuration. */
	configuration = (char *)xmalloc( home_size+conf_size+2);
	strcpy( configuration, home);
	strcpy( configuration+home_size, "/");
	strcpy( configuration+home_size+1, conf_name);

	if (verbose)
		info( "Opening \"%s\".", configuration);
	f = fopen( configuration, "r");
	if (!f)
	{
		if (verbose)
			info_errno_spec( configuration);

		return 0;
	}

	char tmp[ CONF_LINE_MAX];
	while (fgets( tmp, CONF_LINE_MAX, f) != NULL)
	{
		char *t = tmp;
		char *c_name;

		/* Skip spaces. */
		while (*t && isspace( *t))
			t++;

		if (!*t)
			continue;

		/* Find name. */
		c_name = t;
		while (*t && !isspace( *t))
			t++;

		/* Skip empty configuration. */
		if (!*t)
		{
			if (!strcmp( c_name, name))
				return 1;
			continue;
		}

		/* Configuration name test. */
		*t = '\0';
		if (!strcmp( c_name, name))
		{
			bool end = false;
			/* Cycle across parameters. */

			while (!end)
			{
				char *c_conf;

				t++;
				/* Skip spaces. */
				while (*t && isspace( *t))
					t++;
				c_conf = t;
				if (!*t)
					break;
				while (*t && !isspace( *t))
					t++;
				end = !*t;
				*t = '\0';

				param( c_conf);
			}
			
			return 1;
		}
	}

	fclose( f);
	free( configuration);

	return 0;
}


/** Process name or line specification.
 */
static void
param_name( const char *par)

{
	if (!param_find( par))
	{
		if (verbose)
			info( "\"%s\" not found in configuration.", par);
		param_str( par);
	}
}


/** Determines whether parameter is switcheable or not.
 */
static void
param( const char *par)

{
	if (*par == '\0')
		return;

	if (((*par == '-') || (*par == '+') || (*par == '=')) &&
			*(par+1) != '\0')
		param_switch( par);
	else
		param_name( par);
}


/* Going over all parameters and calling param().
 *
 * args		arguments passed to main()
 */
void
parse_param( const char **args)

{
	args++;

	/* No options -> show help. */
	if (*args == '\0')
		parm_help( '-');

	device_name = xstrdup( device_std);

	/* Analyze all options. */
	for (;*args;args++)
		param( *args);
}
